@@ -23,12 +23,12 @@
  *
  */
 
+const Format = imports.format;
 const Gettext = imports.gettext;
 const GLib = imports.gi.GLib;
 
 const Application = imports.ui.application;
 const Path = imports.util.path;
-const Format = imports.util.format;
 const Utils = imports.ui.utils;
 const Tweener = imports.ui.tweener;
 
